@@ -42,8 +42,12 @@ struct _GnomeCupsPPDFile {
 struct _GnomeCupsPrinterDetails {
 	char *printer_name;
 
+	guint attributes_set : 1;
 	guint is_default : 1;
 	guint is_gone : 1;
+	guint is_local : 1;
+
+	guint attributes_request_id;
 
 	/* Option management */
 	guint options_invalid : 1;
@@ -56,6 +60,7 @@ struct _GnomeCupsPrinterDetails {
 	char *info;
 	char *make_and_model;
 	char *device_uri;
+	char *printer_uri;
 	char *state_message;
 	ipp_pstate_t state;
 	int job_count;
@@ -71,6 +76,12 @@ typedef struct {
 	gpointer user_data;
 } AddNotify;
 
+typedef struct {
+	guint id;
+	GnomeCupsPrinterRemovedCallback func;
+	gpointer user_data;
+} RemovedNotify;
+
 static const char *printer_state_strings[] = {
 	N_("Ready"),
 	N_("Printing"),
@@ -91,6 +102,7 @@ static GList *printer_names = NULL;
 static char *default_printer = NULL;
 static GHashTable *printers = NULL;
 static GList *add_notifies = NULL;
+static GList *removed_notifies = NULL;
 static guint signals[LAST_SIGNAL];
 
 GList *
@@ -205,34 +217,22 @@ gnome_cups_printer_get_info (GnomeCupsPrinter *printer)
 	return printer->details->info;
 }
 
-static void
-update_attributes (GnomeCupsPrinter *printer)
-{
 #define MAP_INT(v,a) {if (!g_ascii_strcasecmp (attr->name, (a))) { if ((v) != attr->values[0].integer) { changed = TRUE; } (v) = attr->values[0].integer; }}
 #define MAP_STRING(v,a) {if (!g_ascii_strcasecmp (attr->name, (a))) { if (!v || strcmp (v, attr->values[0].string.text)) { g_free (v); changed = TRUE; (v) = g_strdup (attr->values[0].string.text); }}}
 
-	ipp_t *request;
-	ipp_t *response;	
+static void
+attributes_update_cb (guint id,
+		      const char *path,
+		      ipp_t *response,
+		      GError **error,
+		      gpointer cb_data)
+{
+	GnomeCupsPrinter *printer;
 	ipp_attribute_t *attr;
 	gboolean changed;
-	GError *error = NULL;
-	static const char *attributes[] = {
-	  "printer-state", "queued-job-count",
-	  "printer-location", "printer-info",
-	  "printer-state-message", "device-uri",
-	  "printer-state-reasons", "printer-info",
-	  "printer-make-and-model"
-	};
-	
-	request = gnome_cups_request_new_for_printer (IPP_GET_PRINTER_ATTRIBUTES,
-						      printer->details->printer_name);
-	gnome_cups_request_add_requested_attributes (request, 
-						     IPP_TAG_OPERATION,
-						     G_N_ELEMENTS (attributes),
-						     (char**)attributes);
 
-	response = gnome_cups_request_execute (request, "/", &error);
-	
+	printer = GNOME_CUPS_PRINTER (cb_data);
+
 	changed = FALSE;
 
 	if (!error && response) {
@@ -254,12 +254,13 @@ update_attributes (GnomeCupsPrinter *printer)
 			MAP_STRING (printer->details->state_message, "printer-state-message");
 			MAP_STRING (printer->details->info, "printer-info");
 			MAP_STRING (printer->details->make_and_model, "printer-make-and-model");
+			MAP_STRING (printer->details->printer_uri, "printer-uri-supported");
 		}
 	}
 	ippDelete (response);
-	if (error) {
-		g_error_free (error);
-	}
+	g_clear_error (error);
+
+	printer->details->attributes_set = 1;
 
 	if (changed) {
 		g_free (printer->details->full_state);
@@ -267,8 +268,59 @@ update_attributes (GnomeCupsPrinter *printer)
 		g_signal_emit (printer, signals[ATTRIBUTES_CHANGED], 0);
 	}
 
+	printer->details->attributes_request_id = 0;
+}
+
 #undef MAP_INT
 #undef MAP_STRING
+
+
+static void
+update_attributes (GnomeCupsPrinter *printer)
+{
+	ipp_t *request;
+	char *host;
+	static const char *attributes[] = {
+	  "printer-state", "queued-job-count",
+	  "printer-location", "printer-info",
+	  "printer-state-message", "device-uri",
+	  "printer-state-reasons", "printer-info",
+	  "printer-make-and-model", "printer-uri-supported"
+	};
+
+	if (printer->details->attributes_request_id > 0) {
+		return;
+	}
+	
+	if (printer->details->attributes_set) {
+		request = gnome_cups_request_new_for_printer (IPP_GET_PRINTER_ATTRIBUTES,
+							      printer);
+		gnome_cups_request_add_requested_attributes (request, 
+							     IPP_TAG_OPERATION,
+							     G_N_ELEMENTS (attributes),
+							     (char**)attributes);
+		host = _gnome_cups_printer_get_host (printer);
+	} else {
+		char *printer_uri;
+		request = gnome_cups_request_new (IPP_GET_PRINTER_ATTRIBUTES);
+		printer_uri = g_strdup_printf ("ipp://localhost/printers/%s",
+					       printer->details->printer_name);
+		ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
+			      "printer-uri", NULL, printer_uri);
+		gnome_cups_request_add_requested_attributes (request, 
+							     IPP_TAG_OPERATION,
+							     G_N_ELEMENTS (attributes),
+							     (char**)attributes);
+		g_free (printer_uri);
+		host = NULL;
+	}
+		
+	printer->details->attributes_request_id = 
+		gnome_cups_request_execute_async (request, host, "/",
+						  attributes_update_cb,
+						  g_object_ref (printer),
+						  g_object_unref);
+	g_free (host);
 }
 
 static char *
@@ -308,7 +360,7 @@ get_default (void)
 	cupsFreeDests (num_dests, dests);
 
 	request = gnome_cups_request_new (CUPS_GET_DEFAULT);
-	response = gnome_cups_request_execute (request, "/", &error);
+	response = gnome_cups_request_execute (request, NULL, "/", &error);
 
 	if (error) {
 		ippDelete (response);
@@ -393,6 +445,7 @@ static void
 printer_removed (const char *name)
 {
 	GnomeCupsPrinter *printer;
+	GList *l;
 	
 	printer = gnome_cups_printer_get_existing (name);
 	
@@ -401,6 +454,11 @@ printer_removed (const char *name)
 		g_signal_emit (printer, signals[GONE], 0);
 		g_object_unref (printer);
 	}
+
+	for (l = removed_notifies; l != NULL; l = l->next) {
+		RemovedNotify *notify = l->data;
+		notify->func (name, notify->user_data);
+	}
 }
 
 static gboolean
@@ -425,7 +483,7 @@ get_printer_names (void)
 	
 	request = gnome_cups_request_new (CUPS_GET_PRINTERS);
 	
-	response = gnome_cups_request_execute (request, "/", &error);
+	response = gnome_cups_request_execute (request, NULL, "/", &error);
 
 	if (error) {
 		ippDelete (response);
@@ -501,7 +559,7 @@ update_printers_timeout (gpointer user_data)
 {
 	/* To avoid unneccessary calls during authentication, check
 	 * if a request is currently executing */
-	if (!_gnome_cups_request_is_executing ()) {
+	if (_gnome_cups_outstanding_request_count () == 0) {
 		update_printers ();
 	}
 	
@@ -562,6 +620,43 @@ gnome_cups_printer_new_printer_notify_remove (guint id)
 	set_timeout ();
 }
 
+guint
+gnome_cups_printer_printer_removed_notify_add (GnomeCupsPrinterRemovedCallback cb,
+					       gpointer user_data)
+{
+	guint id = 0;
+	RemovedNotify *notify;
+
+	g_return_val_if_fail (cb != NULL, 0);
+
+	notify = g_new0 (RemovedNotify, 1);
+	
+	notify->id = ++id;
+	notify->func = cb;
+	notify->user_data = user_data;
+	
+	removed_notifies = g_list_append (removed_notifies, notify);
+
+	set_timeout ();
+
+	return notify->id;
+}
+
+void
+gnome_cups_printer_printer_removed_notify_remove (guint id)
+{
+	GList *l;
+	for (l = removed_notifies; l != NULL; l = l->next) {
+		RemovedNotify *notify = l->data;
+		if (notify->id == id) {
+			g_free (notify);
+			removed_notifies = g_list_remove_link (removed_notifies, l);
+		}
+	}
+
+	set_timeout ();
+}
+
 GnomeCupsPrinter *
 gnome_cups_printer_get_existing (const char *printer_name)
 {
@@ -596,10 +691,15 @@ check_exists (const char *printer)
 	GError *error = NULL;
 	ipp_t *request;
 	ipp_t *response;
+	char *printer_uri;
 
-	request = gnome_cups_request_new_for_printer (IPP_GET_PRINTER_ATTRIBUTES,
-						      printer);
-	response = gnome_cups_request_execute (request, "/", &error);
+	request = gnome_cups_request_new (IPP_GET_PRINTER_ATTRIBUTES);
+
+	printer_uri = g_strdup_printf ("ipp://localhost/printers/%s", printer);
+	ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
+		      "printer-uri", NULL, printer_uri);
+	response = gnome_cups_request_execute (request, NULL, "/", &error);
+	g_free (printer_uri);
 
 	ippDelete (response);
 
@@ -649,7 +749,7 @@ gnome_cups_printer_get (const char *printer_name)
 void
 gnome_cups_printer_unref (GnomeCupsPrinter *printer)
 {
-	if (printer)
+	if (printer != NULL)
 		g_object_unref (printer);
 }
 
@@ -669,26 +769,21 @@ gnome_cups_printer_is_gone (GnomeCupsPrinter *printer)
 	return printer->details->is_gone;
 }
 
-static gboolean
-is_local (GnomeCupsPrinter *printer)
+gboolean
+gnome_cups_printer_get_attributes_initialized (GnomeCupsPrinter *printer)
 {
-	int i;
-	static const char *local_uri_schemes[] = {
-		"usb", "serial", "parallel", "epson", "canon", NULL
-	};
-	
-	if (!printer->details->device_uri) {
-		return FALSE;
-	}
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), FALSE);
 
-	for (i = 0; local_uri_schemes[i] != NULL; i++) {
-		if (!strncmp (printer->details->device_uri, 
-			      local_uri_schemes[i],
-			      strlen (local_uri_schemes[i]))) {
-			return TRUE;
-		}
-	}
-	return FALSE;
+	return printer->details->attributes_set;
+}
+
+gboolean
+gnome_cups_printer_get_is_local (GnomeCupsPrinter *printer)
+{
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), FALSE);
+
+	return (printer->details->device_uri != NULL) && 
+		(strcmp (printer->details->device_uri, "") != 0);
 }
 
 void
@@ -699,7 +794,7 @@ gnome_cups_printer_get_icon (GnomeCupsPrinter *printer,
 	g_return_if_fail (GNOME_CUPS_IS_PRINTER (printer));
 	g_return_if_fail (name != NULL);
 
-	*name = is_local (printer) ? g_strdup ("gnome-dev-printer") : g_strdup ("gnome-dev-printer-network");
+	*name = gnome_cups_printer_get_is_local (printer) ? g_strdup ("gnome-dev-printer") : g_strdup ("gnome-dev-printer-network");
 	if (emblems) {
 		*emblems = NULL;
 		if (gnome_cups_printer_get_state (printer) == IPP_PRINTER_STOPPED) {
@@ -712,7 +807,6 @@ gnome_cups_printer_get_icon (GnomeCupsPrinter *printer,
 						  g_strdup ("emblem-default"));
                 }
 	}
-	
 }
 
 ipp_pstate_t
@@ -780,10 +874,25 @@ gnome_cups_printer_get_ppd (GnomeCupsPrinter *printer)
 	return ppd;
 }
 
+const char *
+gnome_cups_printer_get_device_uri (GnomeCupsPrinter const *printer)
+{
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), "");
+	g_return_val_if_fail (printer->details->device_uri, "");
+	return printer->details->device_uri;
+}
+const char *
+gnome_cups_printer_get_uri (GnomeCupsPrinter const *printer)
+{
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), "");
+	g_return_val_if_fail (printer->details->printer_uri, "");
+	return printer->details->printer_uri;
+}
+
 const char *       
 gnome_cups_printer_get_description (GnomeCupsPrinter *printer)
 {
-	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), NULL);
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), "");
 	g_return_val_if_fail (printer->details->description, "");
 
 	return printer->details->description;
@@ -806,10 +915,10 @@ gnome_cups_printer_set_description (GnomeCupsPrinter *printer,
 	}
 
 	request = gnome_cups_request_new_for_printer (CUPS_ADD_PRINTER,
-						      printer->details->printer_name);
+						      printer);
 	ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_TEXT,
 		      "printer-info", NULL, description);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);
 }
@@ -838,10 +947,10 @@ gnome_cups_printer_set_location (GnomeCupsPrinter *printer,
 	}
 
 	request = gnome_cups_request_new_for_printer (CUPS_ADD_PRINTER,
-						      printer->details->printer_name);
+						      printer);
 	ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_TEXT,
 		      "printer-location", NULL, location);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);
 }
@@ -856,8 +965,8 @@ gnome_cups_printer_pause (GnomeCupsPrinter *printer,
 	g_return_if_fail (GNOME_CUPS_IS_PRINTER (printer));
 	
 	request = gnome_cups_request_new_for_printer (IPP_PAUSE_PRINTER,
-						      printer->details->printer_name);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+						      printer);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);
 }
@@ -872,8 +981,8 @@ gnome_cups_printer_resume (GnomeCupsPrinter *printer,
 	g_return_if_fail (GNOME_CUPS_IS_PRINTER (printer));
 	
 	request = gnome_cups_request_new_for_printer (IPP_RESUME_PRINTER,
-						      printer->details->printer_name);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+						      printer);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);	
 }
@@ -888,8 +997,8 @@ gnome_cups_printer_delete (GnomeCupsPrinter *printer,
 	g_return_if_fail (GNOME_CUPS_IS_PRINTER (printer));
 
 	request = gnome_cups_request_new_for_printer (CUPS_DELETE_PRINTER,
-						      printer->details->printer_name);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+						      printer);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 }
 
@@ -1245,6 +1354,9 @@ static void
 gnome_cups_printer_finalize (GObject *object)
 {
 	GnomeCupsPrinter *printer = GNOME_CUPS_PRINTER (object);
+
+	if (printer->details->attributes_request_id > 0)
+		gnome_cups_request_cancel (printer->details->attributes_request_id > 0);
 	
 	if (printer->details->ppd_options) {
 		g_hash_table_destroy (printer->details->ppd_options);
@@ -1484,3 +1596,25 @@ gnome_cups_printer_force_refresh (GnomeCupsPrinter        *printer,
 		printer->details->options_invalid = TRUE;
 	}
 }
+
+gchar *
+_gnome_cups_printer_get_host (GnomeCupsPrinter *printer)
+{
+	gchar *host = NULL;
+	if (printer->details->printer_uri) {
+		gchar *x, *y;
+
+		x = strstr (printer->details->printer_uri, "://");
+
+		if (x) {
+			x += 3;
+			y = strpbrk (x, ":/");
+			if (y)
+				host = g_strndup (x, y - x);
+			else 
+				host = g_strdup (x);
+		}
+	}
+	
+	return host;
+}
